#' Identify and estimate off-target effects in high-throughput RNAi screen
#' 
#' Identify and estimate off-target effects in high-throughput RNAi screen
#' 
#'\tabular{ll}{
#'Package: \tab OffTarget\cr 
#'Type: \tab Package\cr
#'Version: \tab 1.0\cr
#'Date: \tab 2012-10-19\cr
#'License: \tab GPL (>= 2)\cr
#'Depends: \tab glmnet\cr
#'LazyLoad: \tab yes\cr
#'}
#' 
#' @name OffTarget-package
#' @docType package
#' @author Rui Zhong \email{rui.zhong@@utsouthwestern.edu} 
NA

#' Reverse a string
#' 
#' Reserve a given string from the end to the beginning.
#' 
#' @param x a string to be reversed
#' @return reversed string
#' @export
#' @author Rui Zhong \email{rui.zhong@@utsouthwestern.edu} 
StrReverse <- function(x){
  x=as.character(x)
  sapply(lapply(strsplit(x, NULL), rev), paste, collapse="")
} 


#' Reverse a siRNA sequence
#' 
#' Given a siRNA sequence, generate the complementary sequence
#' 
#' @param x a siRNA sequence to be used  
#' @return the complementary sequence of a given siRNA sequence
#' @export
#' @author Rui Zhong \email{rui.zhong@@utsouthwestern.edu} 
DuplexReverse=function(x)
{
  a=as.character(x)
  b=gsub("U","T",a)
  c=gsub("A","U",b)
  d=gsub("T","A",c)
  e=gsub("G","g",d)
  f=gsub("C","G",e)
  g=gsub("g","C",f)
  return(StrReverse(g))
}

#' Extract seed sequence
#' 
#' Given a siRNA sequence, extract the defiend seed sequence
#' 
#' @param x a siRNA sequence to generate the seed sequence  
#' @param start the start position of defined seed sequence  
#' @param end the end position of defined seed sequence
#' @return the seed sequence of a given siRNA sequence by defined region
#' @export
#' @author Rui Zhong \email{rui.zhong@@utsouthwestern.edu} 
SeedExtract=function(x, start=2, end=7){
  x=as.character(x)
  x=substr(x, start, end)
  return(x)
}

#' Test presence of a seed sequence
#' 
#' Given a siRNA sequence, extract the defiend seed sequence
#' 
#' @param dfr a data frame which contain all siRNA seuqences for all pools
#' @param s the seed sequence to be tested 
#' @return a vector of 1 or 0 indicating if a pool contains a seed sequence
#' @export
#' @author Rui Zhong \email{rui.zhong@@utsouthwestern.edu} 
SeedExist=function(dfr, s){
  s=as.character(s)
  return(as.numeric(apply(dfr == s, 1, sum)))
}

#' Seed Analysis
#' 
#' Given RNAi screen sequence and Z score, identify off-target seed families
#' 
#' @param Z normalized Z score from RNAi screening
#' @param data RNAi screen sequence 
#' @param start the start position of defined seed sequence  
#' @param end the end position of defined seed sequence
#' @return object of SeedAnalysis is a list original Z score, original sequence data, a data frame of seed sequence 
#' and a data frame of summary for each seed family.
#' \item{Z}{the input Z score for each pool}
#' 
#' \item{data}{the input sequence data}
#' 
#' \item{seed.table}{a data frame of summary of all seed families with family size and KS test p value 
#' for each seed family}
#' 
#' \item{seed.dfr}{a data frame of seed sequence extraced from data whose region could redefined by user}
#' @export
#' @author Rui Zhong \email{rui.zhong@@utsouthwestern.edu} 
SeedAnalysis=function(Z, data, start=2, end=7){
  
  print("Generating seed.dfr ... ... ")
  seed.dfr=data.frame(data)
  for(i in 1:ncol(seed.dfr)){
    seed.dfr[, i]=as.character(SeedExtract(seed.dfr[, i], start=start, end=end))
  }
  colnames(seed.dfr)=paste("seed", 1:ncol(seed.dfr), sep=".")
  print("Generating seed.table ... ... ")
  seed.table=data.frame(table(unlist(seed.dfr)))
  names(seed.table)=c("seed", "family.size")
  
  seed.table$ks.p.value=rep(NA, length(seed.table$seed))
  y=Z
  for(i in 1:length(seed.table$seed)){
    s=SeedExist(seed.dfr, seed.table$seed[i])
    seed.table$ks.p.value[i]=
      ks.test(y[s > 0], y[s== 0])$p.value
    print(paste("seed ", i))
  }
  
  return(list(Z=Z, data=data, seed.table=seed.table, seed.dfr=seed.dfr))
}

#' Offtarget Analysis
#' 
#' Given a object generated by SeedAnalysis, identify and estimate off-target effcts
#' 
#' @param object a object generated by SeedAnalysis
#' @param s penalty parameter for LASSO regression 
#' @param Intercept indicates in corrected Z score if intercept is included or not. If TURE, then it is included.
#' @return object  of OfftargetAnalysis is a list of corrected Z score, LASSO X matrix, fitting object 
#' and a data frame of summary for the off-target effects of each seed family.
#' \item{fit}{a object of glmnet by performing LASSO regression} 
#' 
#' \item{X}{X matrix in LASSO regression}
#' 
#' \item{offtarget}{a data frame of summary all seed families in a screen with LASSO coefficients,
#'  family size and KS test p value}
#'  
#' \item{Z.corrected}{corrected Z score after removal of off-target effect}
#' @export
#' @author Rui Zhong \email{rui.zhong@@utsouthwestern.edu} 
OfftargetAnalysis=function(object, s=0.001, Intercept=FALSE){
  
  seed.dfr=object$seed.dfr
  seed.table=object$seed.table
  y=object$Z
  
  print("Generating Matrix X ... ... ")
  s.sub=as.character(seed.table$seed)
  x=matrix(0, length(y), length(s.sub))
  colnames(x)=as.character(s.sub)
  
  for(i in 1:length(s.sub)){
    n=as.character(s.sub[i])
    x[, n]=SeedExist(seed.dfr, n)
    print(paste("seed ", i))
  }
  
  print("Fitting ... ... ")
  gc()
  
  fit=glmnet(x, y)
  
  offtarget=data.frame(as.matrix(coef(fit, s=s)))
  offtarget$seed=rownames(offtarget)
  offtarget=offtarget[, c(2, 1)]
  names(offtarget)[2]="coef"
  rownames(offtarget)=NULL
  
  offtarget=merge(offtarget, seed.table, by="seed", all.x=T)
  y.fit=as.numeric(predict(fit, x, s=s))
  
  if(Intercept == T){
    y.corrected=y-y.fit+offtarget[offtarget$seed == '(Intercept)', ]
  }
  else{
    y.corrected=y-y.fit
  }
  
  return(list(fit=fit, X=x, offtarget=offtarget, Z.corrected=y.corrected))
  
}